/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import org.geotools.referencing.GeodeticCalculator;

/**
 * Values derived from the GPS position and time details in IsobusTime.
 * 
 * @author johan
 */
public class GpsDerivedValue {
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.GpsDerivedValue");
    
    private double bearing; // deg
    private double speed; // km/h
    private double distance; // m
    private double interval; // ms
    
    private GpsDerivedValue() {
    }

    public double getBearing() {
        return bearing;
    }

    protected void setBearing(double bearing) {
        this.bearing = bearing;
    }

    public double getSpeed() {
        return speed;
    }

    protected void setSpeed(double speed) {
        this.speed = speed;
    }

    public double getDistance() {
        return distance;
    }

    protected void setDistance(double distance) {
        this.distance = distance;
    }

    public double getInterval() {
        return interval;
    }

    protected void setInterval(double interval) {
        this.interval = interval;
    }
    
    /**
     * Derive values from the GPS details in the input list.
     * 
     * The returned list will always contain null values near the endpoints (number depends on algorithm settings).
     * The output can also contain null values elsewhere in the list, depending on the frequency of unusable GPS measurements.
     * 
     * @param isobusTimes
     * @return list of derived values, same size and indexing as input list.
     */
    protected static List<GpsDerivedValue> calculate(List<IsobusTime> isobusTimes) {
        if (isobusTimes == null) { return null; }
        
        final int segments = 4; // number of line segments included in one vector
        assert segments > 0 && segments % 2 == 0; // spread evenly around center
        
        final int window = 3; // number of points/vectors to include in the moving average; 1 == no smoothing.
        assert window > 0 && window % 2 == 1;
        final int windowHalf = (window -1) / 2; // number of points in the window on one side of the center point

        int lenIso = isobusTimes.size();
        if (lenIso < window + segments - 1) {
            LOGGER.warning("input list too small");
            return null;
        }
        int lenTmp = lenIso - segments;
        
        List<IntermediateValue> intermediateValues = new ArrayList<>(lenTmp);
        List<GpsDerivedValue> gpsDerivedValues = new ArrayList<>(lenIso);
        
        // using GeometryTools would create 2 new calculators each iteration, reuse object.
        GeodeticCalculator calculator = new GeodeticCalculator();

        for (int i = 0; i < lenIso; i++) {
            if (i < segments/2 || i > lenIso - segments/2 - 1) {
                continue;
            }
            IsobusTime lo = isobusTimes.get(i - segments/2);
            IsobusTime hi = isobusTimes.get(i + segments/2);
            if (Positions.isUsable(lo.getPTN()) && Positions.isUsable(hi.getPTN())) {
                calculator.setStartingGeographicPoint(lo.getPTN().getPositionEast().doubleValue(), lo.getPTN().getPositionNorth().doubleValue());
                calculator.setDestinationGeographicPoint(hi.getPTN().getPositionEast().doubleValue(), hi.getPTN().getPositionNorth().doubleValue());
                IntermediateValue tmp = new IntermediateValue();
                double bearing = Math.toRadians(calculator.getAzimuth());
                tmp.sin = Math.sin(bearing);
                tmp.cos = Math.cos(bearing);
                tmp.distance = calculator.getOrthodromicDistance();
                tmp.interval = hi.getTimeStart().getTime() - lo.getTimeStart().getTime();
                intermediateValues.add(tmp);
            } else {
                // NB: this point may still get a derived value later on 
                // if there are enough valid points near it in the moving average window.
                intermediateValues.add(null);
            }
        }
        
        // don't let the default bearing 180 for distance 0 affect the moving averages:
        handleStops(intermediateValues);
        
        final int skip = segments/2 + windowHalf; // number of points near start and end of trajectory which won't get a derived value.
        
        // skip near start:
        for (int i = 0; i < skip; i++) {
            gpsDerivedValues.add(null);
        }
        
        // apply moving averages:
        for (int i = 0; i < lenTmp; i++) {
            if (i < windowHalf || i > lenTmp - windowHalf - 1) {
                continue;
            }
            double sumSin = 0, sumCos = 0, sumDistance = 0, sumInterval = 0, sumSpeed = 0;
            int validCount = 0;
            for (int j = 0; j < window; j++) {
                int offset = i + j - windowHalf;
                IntermediateValue tmp = intermediateValues.get(offset);
                if (tmp != null) {
                    sumSin += tmp.sin;
                    sumCos += tmp.cos;
                    sumDistance += tmp.distance / segments;
                    sumInterval += tmp.interval / segments;
                    sumSpeed += tmp.distance / tmp.interval;
                    validCount++;
                }
            }
            // TODO: might want to set a higher threshold, e.g. windowHalf, ideal case would be equal to window size.
            if (validCount > 0) {
                GpsDerivedValue derived = new GpsDerivedValue();
                double averageAngle = Math.atan2(sumSin, sumCos);
                derived.setBearing(Math.toDegrees(averageAngle));
                derived.setDistance(sumDistance / validCount);
                derived.setInterval(sumInterval / validCount);
                derived.setSpeed(3600 * sumSpeed / validCount); // include UOM conversion: m/ms -> km/h
                gpsDerivedValues.add(derived);
            } else {
                gpsDerivedValues.add(null);
            }
        }
        
        // skip near end:
        for (int i = 0; i < skip; i++) {
            gpsDerivedValues.add(null);
        }
        
        assert gpsDerivedValues.size() == lenIso;
        
        return gpsDerivedValues;
    }
    
    // fill in "missing" bearings before applying averages (the geodetic calculator defaults to 180 degrees for distance 0).
    // fill with the bearing value either directly following or preceding the detected stop, we'll assume there's no rotation while standing still...
    private static void handleStops(List<IntermediateValue> values) {
        int len = values.size();
        boolean moving = true;
        int fillFrom = -1;
        
        // handle stops at start or in the middle of trajectory:
        for (int i = 0; i < len; i++) {
            IntermediateValue value = values.get(i);
            if (value == null) { continue; } // skip (unusable PTN)
            
            // TODO: might use a threshold on distance or speed instead of exactly 0.0
            if (moving && value.distance == 0.0) {
                moving = false;
                fillFrom = i;
            }
            
            if (!moving && value.distance > 0.0) {
                moving = true;
                IntermediateValue filler = value;
                for (int j = fillFrom; j < i; j++) {
                    IntermediateValue toFill = values.get(j);
                    if (toFill != null) {
                        toFill.cos = filler.cos;
                        toFill.sin = filler.sin;
                    }
                }
            }
        }
        // handle stops at end of trajectory:
        if (!moving && fillFrom > 0) {
            IntermediateValue filler = null;
            int preceding = fillFrom - 1;
            do {
                filler = values.get(preceding--);
            } while (filler == null && preceding >= 0);
            
            if (filler != null) {
                for (int j = fillFrom; j < len; j++) {
                    IntermediateValue toFill = values.get(j);
                    if (toFill != null) {
                        toFill.cos = filler.cos;
                        toFill.sin = filler.sin;
                    }
                }
            }
        }
    }
    
    /**
     * Holder for intermediate values over the segments around one point, i.e. for a single vector.
     */
    private static class IntermediateValue {
        // Don't store bearings directly but as unit vector components, so we can safely compute an
        // average for values around the south direction (e.g -179 and 178).
        // average angle = atan2(sumSin, sumCos)
        double sin;
        double cos;
        double distance; // in m
        long interval; // in ms
    }
    
}
