/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.IOException;
import java.util.List;
import java.util.Set;
import nl.cloudfarming.client.geoviewer.BaseLayer;
import nl.cloudfarming.client.geoviewer.RangedLegendPalette;
import nl.cloudfarming.client.imp.ImportSource;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.IsobusGrid;
import nl.cloudfarming.client.util.UOM;
import org.openide.filesystems.FileObject;
import org.openide.loaders.*;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.Lookup;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import org.openide.util.lookup.ProxyLookup;

public class GridDataObject extends MultiDataObject {

    public static final String MIME_TYPE = "application/x-grid";
    public static final String ICON_BASE = "nl/cloudfarming/client/isobus/type/isobus_taskdata_grid.png";
    private final IsobusValueGrid valueGrid;
    private final RangedLegendPalette<Double> palette;
    private final IsobusGrid grid;
    private final FileObject pf;
    private final InstanceContent ic;
    private final Lookup intern;
    private IsobusRaster raster;

    public GridDataObject(FileObject pf, MultiFileLoader loader, List<FileObject> secondary, IsobusGrid grid, IsobusValueGrid valueGrid) throws DataObjectExistsException, IOException {
        super(pf, loader);
        registerEditor(MIME_TYPE, true);
        this.ic = new InstanceContent();
        this.intern = new AbstractLookup(this.ic);
        this.pf = pf;
        this.grid = grid;
        this.valueGrid = valueGrid;

        setSecondaryFiles(secondary);

        DDI ddi = valueGrid.getDDI();
        UOM uom = ddi == null ? null : ddi.getUom();
        double scale = ddi == null ? 1.0 : 1 / ddi.getResolution();
        palette = new RangedLegendPalette<>(0.0, 0.0, uom, scale);

        ic.add(new ImportSource() {

            @Override
            public Set<FileObject> getFiles() {
                return files();
            }
        });

    }

    /**
     * Provides node that should represent this data object.
     *
     * @return the node representation
     * @see DataNode
     */
    @Override
    protected Node createNodeDelegate() {
        if (associateLookup() >= 1) {
            ic.add(getIsobusRaster());
            ic.add(new BaseLayer(palette, pf.getName()));
            DataNode node = new DataNode(this, Children.LEAF, new ProxyLookup(getLookup()));
            node.setIconBaseWithExtension(ICON_BASE);
            return node;
        }
        DataNode dataNode = (DataNode) super.createNodeDelegate();
        return dataNode;
    }

    @Override
    protected int associateLookup() {
        return 1;
    }

    @Override
    public final Lookup getLookup() {
        return new ProxyLookup(super.getLookup(), intern);
    }

    /**
     * get the IsobusRaster which is build from the data in the binary file
     *
     * @return
     */
    public IsobusRaster getIsobusRaster() {
        if (raster == null) {
            if (grid != null) {
                raster = new IsobusRaster(grid, valueGrid, RangedLegendPalette.createDoubleRangeCallback(palette));
            }
        }
        return raster;
    }

    public RangedLegendPalette<Double> getPalette() {
        return palette;
    }

    /**
     * Method to set the secondary files
     *
     * @param secondary the secondary files
     */
    private void setSecondaryFiles(List<FileObject> secondary) {
        for (FileObject fileObject : secondary) {
            registerEntry(fileObject);
        }
    }
}
