/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.awt.Image;
import java.util.Arrays;
import java.util.List;
import nl.cloudfarming.client.geoviewer.BaseLayer;
import nl.cloudfarming.client.geoviewer.Geometrical;
import nl.cloudfarming.client.geoviewer.Layer;
import nl.cloudfarming.client.geoviewer.RangedLegendPalette;
import nl.cloudfarming.client.geoviewer.SimpleLayer;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.util.TimeLogUtil;
import nl.cloudfarming.client.util.UOM;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.ChildFactory;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;
import org.openide.util.lookup.ServiceProvider;

/**
 * <p>Child factory which produces nodes for data in the {@link TimeLogBinaryDataObject}</p>
 * <p>Will search for {@link ServiceProvider} implementations of {@link TimeLogBinaryChildNode}
 * and asks them if they support a {@link TimeLogXMLDataObject}. If they do will
 * ask for a node to add under the main {@link TimeLogBinaryDataObject} Data
 * node.</p>
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
public class TimeLogBinaryChildFactory extends ChildFactory<Node> {

    private final TimeLogBinaryDataObject timeLogBinaryDataObject;

    public TimeLogBinaryChildFactory(TimeLogBinaryDataObject timeLogBinaryDataObject) {
        this.timeLogBinaryDataObject = timeLogBinaryDataObject;
    }

    @Override
    protected boolean createKeys(List<Node> toPopulate) {
        for (TimeLogBinaryChildNodeProvider binaryChildNode : Lookup.getDefault().lookupAll(TimeLogBinaryChildNodeProvider.class)) {
            if (binaryChildNode.supports(timeLogBinaryDataObject)) {
                Node[] nodes = binaryChildNode.getNodes(timeLogBinaryDataObject);
                if (nodes != null) {
                    toPopulate.addAll(Arrays.asList(nodes));
                }
            }
        }
        return true;
    }

    @Override
    protected Node createNodeForKey(Node key) {
        return key;
    }

    /**
     * Child node which provides the trajectory layer
     */
    @ServiceProvider(service = TimeLogBinaryChildNodeProvider.class)
    public static class TrajectoryBinaryChildNode implements TimeLogBinaryChildNodeProvider {

        @Override
        public boolean supports(TimeLogBinaryDataObject timeLogBinaryDataObject) {
            return true;
        }

        @Override
        public Node[] getNodes(TimeLogBinaryDataObject timeLogBinaryDataObject) {
            Geometrical trajectory = timeLogBinaryDataObject.getTrajectory();
            return new TimeLogChildNode[]{new TimeLogChildNode(trajectory, trajectory.getIcon(),
                    trajectory.getTooltipText(), new SimpleLayer(trajectory.getTooltipText()))};
        }
    }

    /**
     * Node which supplies the applied rate or yield grid layer
     */
    @ServiceProvider(service = TimeLogBinaryChildNodeProvider.class)
    public static class RateYieldBinaryChildNode implements TimeLogBinaryChildNodeProvider {

        @Override
        public boolean supports(TimeLogBinaryDataObject timeLogBinaryDataObject) {
            return TimeLogUtil.getAppliedRateYieldDDI(timeLogBinaryDataObject.getXmlHeader().getDDIs()) != null;
        }

        @Override
        public Node[] getNodes(TimeLogBinaryDataObject timeLogBinaryDataObject) {
            // TODO: there can be multiple yield DDIs in a single timelog binary;
            // only the first one found is currently used because the visual panel can only handle one
            // (and before these providers existed that applied to the binary node on the map as well)
            TimeLogGrid grid = timeLogBinaryDataObject.getRateYieldGrid();
            return new TimeLogChildNode[]{new TimeLogChildNode(grid, grid.getIcon(), grid.getTooltipText(),
                    new BaseLayer(timeLogBinaryDataObject.getRateYieldPalette(), grid.getTooltipText()))};
        }
    }

    /**
     * Node which supplies the setpoint rate grid layer.
     * 
     * TODO: control widget order in map when parent node (binary) is shown;
     * setpoint should end up behind applied values, or only be shown when dragged by itself explicitly.
     */
    @ServiceProvider(service = TimeLogBinaryChildNodeProvider.class)
    public static class SetpointRateBinaryChildNode implements TimeLogBinaryChildNodeProvider {

        @Override
        public boolean supports(TimeLogBinaryDataObject timeLogBinaryDataObject) {
            return TimeLogUtil.getSetpointRateDDI(timeLogBinaryDataObject.getXmlHeader().getDDIs()) != null;
        }

        @Override
        public Node[] getNodes(TimeLogBinaryDataObject timeLogBinaryDataObject) {
            DDI ddi = TimeLogUtil.getSetpointRateDDI(timeLogBinaryDataObject.getXmlHeader().getDDIs());
            RangedLegendPalette<Double> palette = new RangedLegendPalette(0, 0, ddi.getUom());
            TimeLogGrid grid = new TimeLogGrid(timeLogBinaryDataObject, ddi, RangedLegendPalette.createDoubleRangeCallback(palette));
            return new TimeLogChildNode[]{new TimeLogChildNode(grid, grid.getIcon(), grid.getTooltipText(), new BaseLayer(palette, grid.getTooltipText()))};
        }
    }
    
    /**
     * Child node for components of the timelog binary file
     */
    private static class TimeLogChildNode extends AbstractNode {

        private String name;
        private Image icon;

        public TimeLogChildNode(Object data, Image icon, String displayName, Layer layer) {
            super(Children.LEAF, new ProxyLookup(Lookups.fixed(data), Lookups.fixed(layer)));
            this.name = displayName;
            this.icon = icon;
        }

        @Override
        public String getDisplayName() {
            return name;
        }

        @Override
        public String getName() {
            return name;
        }

        @Override
        public Image getIcon(int type) {
            return icon;
        }
    }
}
