/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.CoordinateSequence;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.impl.CoordinateArraySequence;
import java.awt.Image;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.*;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import nl.cloudfarming.client.geoviewer.*;
import nl.cloudfarming.client.imp.ImportSource;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.IsobusTaskData;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.isobus.model.util.TimeLogUtil;
import nl.cloudfarming.client.util.Callback;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataNode;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.Exceptions;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;

/**
 * DataObject for TimeLog binary files
 *
 * @author johan
 */
public class TimeLogBinaryDataObject extends MultiDataObject {

    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.TimeLogBinaryDataObject");
    public static final String MIME_TYPE = "application/x-timelog-bin";
    public static final String ICON_BASE = "nl/cloudfarming/client/isobus/type/isobus_taskdata_timelog.png";
    // default date format for datetime columns in the binary; hide date part since it will be constant within a single file.
    public static final String TIMESTAMP_PATTERN = "HH:mm:ss.SSS";
    private final InstanceContent ic;
    private final Lookup intern;
    private final FileObject pf;
    private final List<IsobusTime> isobusTimes;
    private final List<GpsDerivedValue> gpsDerivedValues; // derived from IsobusTime timestamps and GPS coordinates; related by index.
    private final IsobusTaskData isobusTaskData;
    private TimeLogXMLDataObject xmlHeader;
    private Geometrical trajectory;
    private TimeLogGrid rateYieldGrid;
    private RangedLegendPalette<Double> rateYieldPalette;

    protected TimeLogBinaryDataObject(FileObject pf, MultiFileLoader loader, List<FileObject> secondary, List<IsobusTime> isobusTimes, IsobusTaskData isobusTaskData) throws DataObjectExistsException, IOException {
        super(pf, loader);
        registerEditor(MIME_TYPE, true);
        this.ic = new InstanceContent();
        this.intern = new AbstractLookup(this.ic);
        this.pf = pf;
        this.isobusTimes = isobusTimes;
        this.isobusTaskData = isobusTaskData;

        for (FileObject fileObject : secondary) {
            registerEntry(fileObject);
        }

        this.gpsDerivedValues = GpsDerivedValue.calculate(isobusTimes);
        
        ic.add(new ImportSource() {

            @Override
            public Set<FileObject> getFiles() {
                return files();
            }
        });
    }

    @Override
    protected int associateLookup() {
        return 1;
    }

    @Override
    public final Lookup getLookup() {
        return new ProxyLookup(super.getLookup(), intern);
    }

    @Override
    protected Node createNodeDelegate() {
        if (associateLookup() >= 1) {
            DataNode dataNode = new DataNode(this, Children.create(new TimeLogBinaryChildFactory(this), true), new ProxyLookup(getLookup(), Lookups.fixed(new SimpleLayer(pf.getName()))));

            ic.add(new TimeLogGeometrical());

            dataNode.setIconBaseWithExtension(ICON_BASE);
            dataNode.setDisplayName(pf.getName());
            return dataNode;
        }
        return super.createNodeDelegate();
    }

    private Geometry createTrajectoryGeometry() {
        List<Coordinate> coords = new ArrayList<>();
        Coordinate prev = null;
        for (IsobusTime time : isobusTimes) {
            if (Positions.isUsable(time.getPTN())) {
                Coordinate curr = new Coordinate(
                        time.getPTN().getPositionEast().doubleValue(),
                        time.getPTN().getPositionNorth().doubleValue());
                // avoid adjacent duplicates:
                if (!curr.equals(prev)) { coords.add(prev = curr); }
            }
        }
        GeometryFactory geometryFactory = new GeometryFactory();
        Coordinate[] coordArray = coords.toArray(new Coordinate[coords.size()]);
        CoordinateSequence coordSequence = new CoordinateArraySequence(coordArray);
        // linestring must have 0 or >2 points:
        return coords.size() == 1 ? new Point(coordSequence, geometryFactory) : new LineString(coordSequence, geometryFactory);
    }

    public List<IsobusTime> getIsobusTimes() {
        return Collections.unmodifiableList(isobusTimes);
    }

    public List<GpsDerivedValue> getGpsDerivedValues() {
        return gpsDerivedValues == null ? null : Collections.unmodifiableList(gpsDerivedValues);
    }

    public TimeLogXMLDataObject getXmlHeader() {
        if (xmlHeader == null) {
            Pattern timeLogXmlPattern = Pattern.compile(TimeLogXMLFileLoader.PRIMARY_PATTERN);
            for (MultiDataObject.Entry e : secondaryEntries()) {
                FileObject fo = e.getFile();
                if (timeLogXmlPattern.matcher(fo.getNameExt()).find()) {
                    try {
                        DataObject dob = DataObject.find(fo);
                        xmlHeader = (TimeLogXMLDataObject) dob;
                    } catch (DataObjectNotFoundException | ClassCastException ex) {
                        Exceptions.printStackTrace(ex);
                    }
                }
            }
        }
        return xmlHeader;
    }
    
    public IsobusTaskData getTaskData() {
        return isobusTaskData;
    }

    /**
     * Get the trajectory geometrical
     *
     * @return
     */
    public Geometrical getTrajectory() {
        if (trajectory == null) {
            trajectory = new TrajectoryGeometrical(createTrajectoryGeometry());
        }
        return trajectory;
    }

    /**
     * get the applied rate/yield Grid
     *
     * @return
     */
    public TimeLogGrid getRateYieldGrid() {
        RangedLegendPalette<Double> palette = getRateYieldPalette();
        if (rateYieldGrid == null &&  palette != null) {
            DDI ddi = TimeLogUtil.getAppliedRateYieldDDI(getXmlHeader().getDDIs());
            rateYieldGrid = new TimeLogGrid(this, ddi, RangedLegendPalette.createDoubleRangeCallback(palette));
        }
        return rateYieldGrid;
    }

    /**
     * Get the rate yield palette Returns null if no plottable ddi can be found
     *
     * @return
     */
    public RangedLegendPalette<Double> getRateYieldPalette() {
        if (rateYieldPalette == null) {
            DDI ddiToPlot = TimeLogUtil.getAppliedRateYieldDDI(getXmlHeader().getDDIs());
            if (ddiToPlot != null) {
                //start with dummy values, set actual range via callback when raster is generated:
                rateYieldPalette = new RangedLegendPalette<>(0.0, 0.0, ddiToPlot.getUom());
            }
        }
        return rateYieldPalette;
    }

    /**
     * Simple geometrical which uses the trajectory to get it's centroid
     */
    private class TimeLogGeometrical implements Geometrical {

        Geometry geometry;
        
        public TimeLogGeometrical() {
            this.geometry = createTrajectoryGeometry();
        }

        @Override
        public Geometry getGeometry() {
            return null;
        }

        @Override
        public Point getCentroid() {
            return (geometry == null || !geometry.isValid()) ? null : geometry.getCentroid();
        }

        @Override
        public Geometry getBoundingBox() {
            return getCentroid();
        }

        @Override
        public Object getRenderObject(Envelope envelope) {
            return null;
        }

        @Override
        public String getTooltipText() {
            return "";
        }

        @Override
        public Image getIcon() {
            return null;
        }

        @Override
        public String getIconLabel() {
            return "";
        }

        @Override
        public void addPropertyChangeListener(PropertyChangeListener listener) {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        @Override
        public void removePropertyChangeListener(PropertyChangeListener listener) {
            throw new UnsupportedOperationException("Not supported yet.");
        }
    }

    // geometrical to show the trajectory on the map and in preview panels
    @NbBundle.Messages({"TrajectoryGeometrical trajectory display name=Trajectory"})
    private class TrajectoryGeometrical implements Geometrical {

        private final Geometry geometry;
        private final String TRAJECTORY_DISPLAYNAME = Bundle.TrajectoryGeometrical_trajectory_display_name();
        private final String TRAJECTORY_ICON_BASE = "nl/cloudfarming/client/isobus/type/trajectory.png";

        public TrajectoryGeometrical(Geometry geometry) {
            this.geometry = geometry;
        }

        @Override
        public Geometry getGeometry() {
            return geometry;
        }

        @Override
        public Point getCentroid() {
            return (geometry == null || !geometry.isValid()) ? null : geometry.getCentroid();
        }

        @Override
        public Geometry getBoundingBox() {
            if (geometry == null || !geometry.isValid()) { return null; }
            Geometry envelope = geometry.getEnvelope();
            return envelope.getNumPoints() < 4 ? null : envelope;
        }

        @Override
        public Object getRenderObject(Envelope envelope) {
            return null;
        }

        @Override
        public String getTooltipText() {
            return TRAJECTORY_DISPLAYNAME;
        }

        @Override
        public Image getIcon() {
            return ImageUtilities.loadImage(TRAJECTORY_ICON_BASE);
        }

        @Override
        public String getIconLabel() {
            return TRAJECTORY_DISPLAYNAME;
        }

        @Override
        public void addPropertyChangeListener(PropertyChangeListener listener) {
        }

        @Override
        public void removePropertyChangeListener(PropertyChangeListener listener) {
        }
    }

    // geographical to show the applied rate / yield maps
    public class RateYieldGrid extends TimeLogGrid {

        public RateYieldGrid(final TimeLogBinaryDataObject binary, final DDI ddi, final Callback<double[]> rangeCallback) {
            super(binary, ddi, rangeCallback);
        }

    }
}
