/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.xml.parsers.ParserConfigurationException;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionReferences;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.MIMEResolver;
import org.openide.loaders.*;
import org.openide.loaders.MultiDataObject.Entry;
import org.openide.util.Exceptions;
import org.openide.util.NbBundle;
import org.xml.sax.SAXException;

/**
 * File loader for Time log header xml file
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
@DataObject.Registration(mimeType = TimeLogXMLDataObject.MIME_TYPE,
iconBase = TimeLogXMLDataObject.ICON_BASE,
displayName = "#LBL_TimeLogXML_LOADER",
position = 300)
@MIMEResolver.Registration(displayName = "#LBL_TimeLogXML_LOADER",
resource = "TimeLogXMLResolver.xml")
@ActionReferences(value = {
    @ActionReference(id =
    @ActionID(category = "System", id = "org.openide.actions.OpenAction"), path = "Loaders/application/x-timelog-xml/Actions", position = 100, separatorAfter = 200),
    @ActionReference(id =
    @ActionID(category = "System", id = "org.openide.actions.ToolsAction"), path = "Loaders/application/x-timelog-xml/Actions", position = 1300)})
@NbBundle.Messages({
    "LBL_TimeLogXML_LOADER=Files of TimeLog xml"
})
public class TimeLogXMLFileLoader extends MultiFileLoader {
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.TimeLogXMLFileLoader");

    public static final String PRIMARY_PATTERN = "(?i)tlg[0-9]{5}\\.xml";
    // secondary files and dependencies:
    private static final String TASKDATA_PATTERN = TaskDataFileLoader.PRIMARY_PATTERN;
    private static final String TIMELOG_BIN_PATTERN = TimeLogBinaryFileLoader.PRIMARY_PATTERN;

    public TimeLogXMLFileLoader() {
        super(TimeLogXMLDataObject.class.getName());
    }

    @Override
    protected FileObject findPrimaryFile(FileObject fo) {
        // never recognize folders.
        if (fo.isFolder()) {
            return null;
        }

        Pattern timeLogXmlPattern = Pattern.compile(PRIMARY_PATTERN);

        if (timeLogXmlPattern.matcher(fo.getNameExt()).find()) {
            return fo;
        }
        // not recognized:
        return null;
    }

    /**
     * 
     *
     * @param primaryFile
     * @return
     * @throws DataObjectExistsException
     * @throws IOException
     */
    @Override
    protected MultiDataObject createMultiObject(FileObject primaryFile) throws DataObjectExistsException, IOException {
        List<FileObject> secondary = new ArrayList<>();
        Pattern timeLogBinPattern = Pattern.compile(TIMELOG_BIN_PATTERN);
        Pattern taskDataPattern = Pattern.compile(TASKDATA_PATTERN);
        FileObject taskDataFile = null;
        InvalidDataObject.Detail invalidDataDetail; // userfriendly message to be used in tooltip or error panel.
        
        for (FileObject child : primaryFile.getParent().getChildren()) {
            if (!child.equals(primaryFile)) {
                if (taskDataPattern.matcher(child.getNameExt()).find()) {
                    // required to load data but don't claim it as a secondary file:
                    taskDataFile = child;
                } else if (timeLogBinPattern.matcher(child.getNameExt()).find()
                        && child.getName().equalsIgnoreCase(primaryFile.getName())) {
                    secondary.add(child);
                }
            }
        }
        
        if (taskDataFile == null) {
            LOGGER.log(Level.SEVERE, "No taskdata.xml file found for {0}", primaryFile.getNameExt());
            invalidDataDetail = InvalidDataObject.detail(Bundle.InvalidDataObject_error_missing_dependency("taskdata.xml"));
        } else {
            
            try {
                DataObject dataObject = DataObject.find(taskDataFile);
                if (dataObject instanceof TaskDataObject) {
                    TaskDataObject isobusDataObject = (TaskDataObject) dataObject;
                    TimeLogXMLReader xMLReader = new TimeLogXMLReader(primaryFile.getInputStream());
                    Map<TimeFieldEnum, String> timeMap = xMLReader.getTimeElement();
                    Map<PositionFieldEnum, String> positionMap = xMLReader.getPositionElement();
                    List<Map<DatalogValueFieldEnum, String>> dataLogMap = xMLReader.getDataLogValues();
                    return new TimeLogXMLDataObject(primaryFile, this, secondary, timeMap, positionMap, dataLogMap, isobusDataObject.getIsobusTaskData());
                } else {
                    invalidDataDetail = InvalidDataObject.detail(Bundle.InvalidDataObject_error_invalid_dependency("taskdata.xml"));
                }
            } catch (ParserConfigurationException | SAXException | FileNotFoundException ex) {
                Exceptions.attachSeverity(ex, Level.INFO);
                Exceptions.printStackTrace(ex);
                LOGGER.log(Level.SEVERE, "Exception while loading timelog xml header: {0}", ex.toString());
                invalidDataDetail = InvalidDataObject.detail("Exception while loading timelog XML data", ex);
            }
        }
        return new InvalidDataObject(primaryFile, this, TimeLogXMLDataObject.ICON_BASE, invalidDataDetail);
    }

    @Override
    protected Entry createPrimaryEntry(MultiDataObject obj, FileObject primaryFile) {
        Entry entry = new FileEntry(obj, primaryFile);
        LOGGER.log(Level.FINEST, "Creating primary entry for {0}", primaryFile.getNameExt());
        return entry;
    }

    @Override
    protected Entry createSecondaryEntry(MultiDataObject obj, FileObject secondaryFile) {
        FileEntry entry = new FileEntry(obj, secondaryFile);
        LOGGER.log(Level.FINEST, "Creating secondary entry for {0}", secondaryFile.getNameExt());
        return entry;
    }
}
